--! @file    	adis16209_control.vhd
--! @brief   	VHDL-Modul, first processed file by DoxyGen 1.8.10
--! @details \n
--! Der ADIS16209 erfasst selbstständig Messwerte und stellt diese in Registern im Raw-Format über ein SPI-Interface zur Verfügung. \n
--! Der ADIS wird in einer fixen Konfiguration betrieben:
--! - es werden nur diejenigen Register verarbeitet, die für den Tumbler relevant sind
--! - nach Power-Up wird die Produkt-ID abgefragt
--! - der aktuelle Status wird über das STATUS-Register ausgelesen (mit anschließendem Clearing)
--!
--! <b>Wesentliche Funktionen</b>\n
--! - 2-stufige Synchronisation der IN-Ports mit dem Systemtakt, um ein stabiles VHDL-Verhalten bei Leseoperationen zu gewährleisten (Prozess "adis_port_sync")
--! - reduzierter Systemclock für SPI-Interface, SPI-SCLK = 500kHz, Erzeugung des asynchr. Enable-Signals (Prozess "spi_adis_sclk_x_tel")
--! - Messwert-Aufbereitung, Mittelwertbildung und synchronisierte Port-Ausgabe (Prozess "outport_sync") \n
--!		- Die 16bit-ADIS-Raw-Messdaten werden auf das reale Datenformat (14 bits, ohne ND/EA-bits) beschnitten und bei Bedarf im 2's complement formatiert. \n
--!		- Bei aktiviertem Averaging-QSys-Parameter werden die Messwerte dann einer gleitenden 8-Level-Mittelwertbildung unterzogen. \n
--!
--! - Gesamtsteuerung (Prozess "adis_main_statemachine") \n
--! 		- Abschnitt Power-Up, Initialisierung, Statusabfrage und Parameterkonfiguration \n
--! 		- Abschnitt RX SINGLE Data Word from ADIS \n
--!
--!		- Im Messbetrieb liefert der ADIS über DIO1 ein "DataReady"-Signal am Ende jeder kompletten Messwerterfassung und erlaubt das Auslesen aller Messwert-Register. \n
--!
--!
--! \n@note
--! - Das beobachtete Verhalten des ADIS16209 weicht manchmal vom Datenblatt (bzw. der Erwartungshaltung) ab, Details siehe kommentiertes Datenblatt.
--! - Es wird hier auch bewusst auf eine Nios-Steuerung über ein Command-Port verzichtet, da der ADIS in einer fixen Konfiguration läuft.
--!   Bei Bedarf muss ggf. das Command-Port hinzugefügt bzw. aktiviert werden.
--!
--! \n@verbatim
--! Modul-Interface-Register
--! Die Messwerte werden über folgende Ausgaberegister zur Verfügung gestellt (in dieser Reihenfolge, Inhalt siehe Datenblatt, bereits formatiert für 32bit-Nios-Lesen)
--! 	#1	XACCL_OUT	: real 14bit 2's complement format
--! 	#2	YACCL_OUT	: real 14bit 2's complement format
--!
--! Production-ID-Register
--! #6 PROD_ID          : real 16bit binary format, Silizium-ID 0x3F51 (=16209)
--!
--! Status-Register
--! #7 STATUS           : real 16bit binary format
--!
--! zusätzliches eigenes Fehler-Flag-Register
--! #8 ERROR            : real 16bit binary format, Mapping der Fehlerflags (EA-bit) der einzelnen Datenblöcke (Löschen nur über Reset)
--!  _________________________________________________________________________________________________________________________________
--! |     15|       |       |       |       |       |       |       8||      7|      6|      5|      4|      3|      2|      1|      0|
--! |       |       |       |       |       |       |       |        ||       |       |       |       |       |       |       |       |
--! |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |  ---   ||  ---  |  ---  |   #5  |  #4   |  #3   |  #2   |  #1   |  ---  |
--! |_______|_______|_______|_______|_______|_______|_______|________||_______|_______|_______|_______|_______|_______|_______|_______|
--!
--! @endverbatim \n
--!
--! compilation : QuartusII, V15.0.1, Build 150 \n
--! simulation : ModelSim, V10.3d rev. 2014.10 \n
--! 
--! @author  JMayr
--! @version V1.0
--! @date    02.09.2015
--! 
--! \n@par History:
--! @verbatim
--! V1.0	first release
--!		JMayr, 02.09.2015
--! @endverbatim
--!
--! \n@todo  nichts eingetragen
--!
--! @bug  nichts eingetragen
--!
--! @brief
--! <h2><center>&copy; COPYRIGHT 2011 UniBwM ETTI WE 4</center></h2>
--!
--! @defgroup VHDL VHDL

--! Use Standard Library
LIBRARY IEEE;

--! Use Standard Logic Library
USE IEEE.STD_LOGIC_1164.ALL;

--! Use Standard Arithmetic Library
USE IEEE.STD_LOGIC_ARITH.ALL;     -- für conv functions erforderlich

-- --! Use Standard Logic Unsigned Library
-- USE IEEE.STD_LOGIC_unsigned.ALL;  -- für conv functions erforderlich

--! Use Standard Logic Signed Library
USE IEEE.STD_LOGIC_signed.ALL;  -- alternativ zur unsigned-Lib...

-- --! Use Numeric Library
-- USE ieee.numeric_bit.ALL;
-- --! Use Numeric Library
-- USE ieee.numeric_std.ALL;

-- LIBRARY modelsim_lib;
-- use modelsim_lib.util.all;

--! @ingroup VHDL
--! @brief Ansteuerung ADIS16209
ENTITY adis16209_control is
GENERIC(
	spi_serial_clock_div_g 			: integer := 50; 	--! Teilerfaktor für operativen SPI-SCLK-Betrieb mit 500kHz (= 2µs) : 50MHz / (2 x 500kHz) = 50 => 1MHz reduz. Clock (wg. adis_sclk_x_tel_s)
	smpld_prd_timebase_setting_g	: integer := 0;	--! Bit SMPL_PRD[7] 		---> '0' = Timebase 244,14µs
	smpld_prd_increment_setting_g : integer := 5;	--! Bits SMPL_PRD[6:0] 	---> '0000101' = Increment Setting Ns = 5
	avg_cnt_setting_g					: integer := 8;	--! Bits AVG_CNT[3:0]	---> '1000' = (Defaultwert)
	enable_data_averaging_g			: integer := 1		--! Averaging 8-level-Buffer für Messwerte: enable = 1, disable = 0	(wg. QSys nur Integer-Angabe)
	);
PORT(
	COMMAND		: in std_logic_vector(15 downto 0);	--! 16-bit-Command-Port (hier nicht verwendet, weil ADIS in fester Konfiguration betrieben wird)
	
	-- ADIS16209-Daten- und Messwerte, für NIOS-Lesen bereits in 32bit-Standard-Wortbreite, ggf. bereits im 2's-Complement Format
	XACCL_OUT	: out std_logic_vector(31 downto 0);	--! real 14bit 2's complement format
	YACCL_OUT	: out std_logic_vector(31 downto 0);	--! real 14bit 2's complement format
	XINCL_OUT	: out std_logic_vector(31 downto 0);	--! real 14bit 2's complement format

	--eigenes Fehlerregister
	ERROR			: out std_logic_vector(31 downto 0);	--! real 16bit binary format, Mapping der Fehlerflags (EA-bit) der einzelnen Datenblöcke
	
	-- ADIS16209-Interface
	ADIS16209_CS_n			: OUT std_logic;					--! low-aktives ADIS16209-Chip-Select (In-Pin "CS")
	ADIS16209_SCLK			: OUT std_logic; 					--! Clock zum ADIS16209 (In-Pin "SCLK")

	CLK			: IN std_logic; 								--! System-Clock
	RESET_n		: IN std_logic									--! System-Reset
	);
END adis16209_control;


--! @brief Architecture für adis16209_control
ARCHITECTURE arch_adis16209_control of adis16209_control is

signal adis_sclk_x_tel_s			: std_logic;	--! Unterteilung einer SCLK-Periode in x Teile, wg. sauberem Versatz von Daten- und Taktflanken, wirkt als asynchr. enable-Signal in den Prozessen

--ADIS16209-Pin-Signale
signal adis_cs_n_s						: std_logic;	--! low-aktives ADIS16209-Chip-Select
signal adis_sclk_s						: std_logic;	--! Clock zum ADIS16209
signal adis_din_s							: std_logic;	--! serielle Daten zum ADIS16209

--definition main_state type
type main_state_type is ( 														--!States für Steuerung und Ablauf
	INIT_A_ST, INIT_B_ST, INIT_C_ST, INIT_D_ST, INIT_E_ST, INIT_F_ST, INIT_G_ST, INIT_H_ST, INIT_I_ST, INIT_J_ST, INIT_K_ST, WAIT_DATA_READY_ST,
	GET_DATA_WORD_1_ST, GET_DATA_WORD_2_ST, GET_DATA_WORD_3_ST, GET_DATA_WORD_4_ST, GET_DATA_WORD_5_ST, GET_DATA_WORD_6_ST,
	GET_ALL_REGISTER_DATA_1_ST, GET_ALL_REGISTER_DATA_2_ST, GET_ALL_REGISTER_DATA_3_ST, GET_ALL_REGISTER_DATA_4_ST, GET_ALL_REGISTER_DATA_5_ST, GET_ALL_REGISTER_DATA_6_ST, GET_ALL_REGISTER_DATA_7_ST,
	TX_TO_ADIS_1_ST, TX_TO_ADIS_2_ST, TX_TO_ADIS_3_ST, TX_TO_ADIS_4_ST, TX_TO_ADIS_5_ST,
	ERROR_ST
	);

--Definition eines Averaging 8-level buffer types
type type_averaging_buffer is array (7 downto 0) of integer;		--! Datentyp für zusätzliche Messwert-Averaging-Funktion mit fixem 16-bit-Buffer für 8 Durchläufe (keine integer-Range-Angabe wg. Vorzeichen)
signal averaging_buffer_xaccl_out_s		: type_averaging_buffer;	--! Averaging-Buffer für XACCL_OUT
signal averaging_buffer_yaccl_out_s		: type_averaging_buffer;	--! Averaging-Buffer für YACCL_OUT


--! @brief VHDL-Procedure: It seems that this causes problems in documenting VHDL-Generics
--! @details 	The first processed file in the DoxyGen file list shows all Generics. \n
--!				But the second file does skip the Generics !
--!				Just put the procedure statements in the first processed file into comments ---> Generics in both files are shown
--! @param  type_of_result_v : aktueller zu mittelnder ADIS-Messwert
--! @param  averaging_buffer_<adis-register_out_s> : jeweils der Average-Buffer für den aktuell zu mittelnden Messwert
--! @retval local_averaged_result_v : errechneter Mittelwert
procedure proc_result_averaging (variable type_of_result_v					: integer;
											variable bitmuster_v							: in std_logic_vector;
											signal averaging_buffer_xaccl_out_s		: inout type_averaging_buffer;	--case type_of_result_v = 1
											signal averaging_buffer_yaccl_out_s		: inout type_averaging_buffer;	--case type_of_result_v = 2
											signal averaging_buffer_xincl_out_s		: inout type_averaging_buffer;	--case type_of_result_v = 3
											signal averaging_buffer_yincl_out_s		: inout type_averaging_buffer;	--case type_of_result_v = 4
											signal averaging_buffer_rot_out_s		: inout type_averaging_buffer;	--case type_of_result_v = 5
											variable local_averaged_result_v			: out integer) is

begin
case type_of_result_v is
	when 1 => 		--xaccl_out
		averaging_buffer_xaccl_out_s(7) 	<= averaging_buffer_xaccl_out_s(6);
		averaging_buffer_xaccl_out_s(6) 	<= averaging_buffer_xaccl_out_s(5);
		averaging_buffer_xaccl_out_s(5) 	<= averaging_buffer_xaccl_out_s(4);
		averaging_buffer_xaccl_out_s(4) 	<= averaging_buffer_xaccl_out_s(3);
		averaging_buffer_xaccl_out_s(3) 	<= averaging_buffer_xaccl_out_s(2);
		averaging_buffer_xaccl_out_s(2) 	<= averaging_buffer_xaccl_out_s(1);
		averaging_buffer_xaccl_out_s(1) 	<= averaging_buffer_xaccl_out_s(0);
		averaging_buffer_xaccl_out_s(0) 	<= conv_integer(bitmuster_v);
		local_averaged_result_v := (	averaging_buffer_xaccl_out_s(0) +
												averaging_buffer_xaccl_out_s(1) +
												averaging_buffer_xaccl_out_s(2) +
												averaging_buffer_xaccl_out_s(3) +
												averaging_buffer_xaccl_out_s(4) +
												averaging_buffer_xaccl_out_s(5) +
												averaging_buffer_xaccl_out_s(6) +
												averaging_buffer_xaccl_out_s(7) ) / 8;

	when others => null;
end case;
end procedure proc_result_averaging;

--Signale für eingelesene ADIS Output Data Register im Raw-Format (alles 16bit, inkl. ND und EA bits, ggf. im 2's complement format)
signal xaccl_out_raw_reg_s		: std_logic_vector(15 downto 0);	--! ADIS Output Data Register im Raw-Format (16bit, inkl. ND und EA bit, real 14bit, 2's complement format)
signal yaccl_out_raw_reg_s		: std_logic_vector(15 downto 0);	--! ADIS Output Data Register im Raw-Format (16bit, inkl. ND und EA bit, real 14bit, 2's complement format)

--ADIS-Config-Konstanten
--constant powerup_div_factor_c				:integer := 11;		-- ******************* Vorsicht: testweise nur zur Simulation ************************
constant powerup_div_factor_c					:integer := 2**21;	--! Teilerfaktor für die Power-Up-Wartezeit, mind. 1460ms (s. Datenblatt S.4, reduz. Systemclock 1MHz(1µs) * 2**21 => ca. 2s)

constant prod_id_reg_read_mask_c				:std_logic_vector(15 DOWNTO 0) := "0100101000000000";	--! Reg. PROD_ID, read lower/upper Byte, ADR = 0x4A, Bitmuster "0_1001010_0000_0000", Silizium-ID 0x3F51 = 16209


--**************************************************************************
--**************************************************************************
begin	--of architecture ****************************************************
--**************************************************************************
--**************************************************************************

--HW-ADIS16209-Signal-to-Port Assignment
ADIS16209_DIN 	<= adis_din_s;
ADIS16209_CS_n <= adis_cs_n_s;
ADIS16209_SCLK <= adis_sclk_s;


--! @brief Messwert-Aufbereitung, Mittelwertbildung und synchronisierte Port-Ausgabe
--! @details 	Die 16bit-ADIS-Raw-Messdaten werden auf das reale Datenformat (14 bits, ohne ND/EA-bits) beschnitten und bei Bedarf im 2's complement formatiert. \n
--!				Bei aktiviertem Averaging-QSys-Parameter werden die Messwerte dann einer gleitenden 8-Level-Mittelwertbildung unterzogen. \n
--!				Die Ausgabe erfolgt synchron mit dem "reduzierten" Systemtakt von 1MHz somit als "signed integer"-Zahlenwert.
--!				Eine evtl. Umrechnung dieser Zahlenwerte in andere physikalische Einheiten (z.B. "g", "V" oder "°/sec") erfolgt in der Applikationsoftware.
outport_sync: process  (RESET_n, CLK, adis_sclk_x_tel_s,
								xaccl_out_raw_reg_s, yaccl_out_raw_reg_s, xincl_out_raw_reg_s, yincl_out_raw_reg_s, rot_out_raw_reg_s,
								status_reg_s, error_reg_s)
							
variable bitmuster_v			: std_logic_vector(31 downto 0);	--Zwischenvariable zum temporären "Bit-Mixen"
variable type_of_result_v 	: integer range 0 to 15;			--Kennzeichnung der Art des Messwertes, damit in Average-Procedure die richtige Mittelwertbildung erfolgt
variable averaged_result_v	: integer;								--errechnetes Mittelwertergebnis 

begin
if RESET_n = '0' then
	bitmuster_v := x"00000000";
elsif adis_sclk_x_tel_s = '0' then           --asynchr. enable signal für "reduzierten Systemclock"
	null;
elsif CLK'EVENT and CLK = '1' then				--reduzierter Clock, de facto 1 MHz
	--für XACCL_OUT Register, real 14bit 2's complement format
	if (xaccl_out_raw_reg_s(13) = '0') then	--pos. Zahl
	 	bitmuster_v := x"00000000";				--alles auf '0' setzen
	else													--neg. Zahl
	 	bitmuster_v := x"FFFFFFFF";				--alles auf '1' setzen wg. 2's complement
	end if;
	bitmuster_v(13 downto 0) := xaccl_out_raw_reg_s(13 downto 0);	--nur den realen Zahlenwert einkopieren
	type_of_result_v := 1;							--Kennzeichnung für Mittelwert-Prozedur setzen
	if (enable_data_averaging_g = 1) then
		proc_result_averaging (	type_of_result_v,
										bitmuster_v,
										averaging_buffer_xaccl_out_s,
										averaging_buffer_yaccl_out_s,
										averaging_buffer_xincl_out_s,
										averaging_buffer_yincl_out_s,
										averaging_buffer_rot_out_s,
										averaged_result_v);
		XACCL_OUT <= conv_std_logic_vector(averaged_result_v, 32);		--gemittelten Datenwert auf Port legen
	else
		XACCL_OUT <= bitmuster_v;		--einzelnen (nicht gemittelten) Datenwert auf Port legen
	end if;
		
	--für YACCL_OUT Register, real 14bit 2's complement format
	if (yaccl_out_raw_reg_s(13) = '0') then	--pos. Zahl
	 	bitmuster_v := x"00000000";				--alles auf '0' setzen
	else													--neg. Zahl
	 	bitmuster_v := x"FFFFFFFF";				--alles auf '1' setzen wg. 2's complement
	end if;
	bitmuster_v(13 downto 0) := yaccl_out_raw_reg_s(13 downto 0);	--nur den realen Zahlenwert einkopieren
	type_of_result_v := 2;							--Kennzeichnung für Mittelwert-Prozedur setzen
	if (enable_data_averaging_g = 1) then
		proc_result_averaging (	type_of_result_v,
										bitmuster_v,
										averaging_buffer_xaccl_out_s,
										averaging_buffer_yaccl_out_s,
										averaging_buffer_xincl_out_s,
										averaging_buffer_yincl_out_s,
										averaging_buffer_rot_out_s,
										averaged_result_v);
		YACCL_OUT <= conv_std_logic_vector(averaged_result_v, 32);		--gemittelten Datenwert auf Port legen
	else
		YACCL_OUT <= bitmuster_v;		--einzelnen (nicht gemittelten) Datenwert auf Port legen
	end if;

	--für XINCL_OUT Register, real 14bit 2's complement format
	if (xincl_out_raw_reg_s(13) = '0') then	--pos. Zahl
	 	bitmuster_v := x"00000000";				--alles auf '0' setzen
	else													--neg. Zahl
	 	bitmuster_v := x"FFFFFFFF";				--alles auf '1' setzen wg. 2's complement
	end if;
	bitmuster_v(13 downto 0) := xincl_out_raw_reg_s(13 downto 0);	--nur den realen Zahlenwert einkopieren
	type_of_result_v := 3;							--Kennzeichnung für Mittelwert-Prozedur setzen
	if (enable_data_averaging_g = 1) then
		proc_result_averaging (	type_of_result_v,
										bitmuster_v,
										averaging_buffer_xaccl_out_s,
										averaging_buffer_yaccl_out_s,
										averaging_buffer_xincl_out_s,
										averaging_buffer_yincl_out_s,
										averaging_buffer_rot_out_s,
										averaged_result_v);
		XINCL_OUT <= conv_std_logic_vector(averaged_result_v, 32);		--gemittelten Datenwert auf Port legen
	else
		XINCL_OUT <= bitmuster_v;		--einzelnen (nicht gemittelten) Datenwert auf Port legen
	end if;

	--für YINCL_OUT Register, real 14bit 2's complement format
	if (yincl_out_raw_reg_s(13) = '0') then	--pos. Zahl
	 	bitmuster_v := x"00000000";				--alles auf '0' setzen
	else													--neg. Zahl
	 	bitmuster_v := x"FFFFFFFF";				--alles auf '1' setzen wg. 2's complement
	end if;
	bitmuster_v(13 downto 0) := yincl_out_raw_reg_s(13 downto 0);	--nur den realen Zahlenwert einkopieren
	type_of_result_v := 4;							--Kennzeichnung für Mittelwert-Prozedur setzen
	if (enable_data_averaging_g = 1) then
		proc_result_averaging (	type_of_result_v,
										bitmuster_v,
										averaging_buffer_xaccl_out_s,
										averaging_buffer_yaccl_out_s,
										averaging_buffer_xincl_out_s,
										averaging_buffer_yincl_out_s,
										averaging_buffer_rot_out_s,
										averaged_result_v);
		YINCL_OUT <= conv_std_logic_vector(averaged_result_v, 32);		--gemittelten Datenwert auf Port legen
	else
		YINCL_OUT <= bitmuster_v;		--einzelnen (nicht gemittelten) Datenwert auf Port legen
	end if;

	--für ROT_OUT Register, real 14bit 2's complement format
	if (rot_out_raw_reg_s(13) = '0') then		--pos. Zahl
	 	bitmuster_v := x"00000000";				--alles auf '0' setzen
	else													--neg. Zahl
	 	bitmuster_v := x"FFFFFFFF";				--alles auf '1' setzen wg. 2's complement
	end if;
	bitmuster_v(13 downto 0) := rot_out_raw_reg_s(13 downto 0);	--nur den realen Zahlenwert einkopieren
	type_of_result_v := 5;							--Kennzeichnung für Mittelwert-Prozedur setzen
	if (enable_data_averaging_g = 1) then
		proc_result_averaging (	type_of_result_v,
										bitmuster_v,
										averaging_buffer_xaccl_out_s,
										averaging_buffer_yaccl_out_s,
										averaging_buffer_xincl_out_s,
										averaging_buffer_yincl_out_s,
										averaging_buffer_rot_out_s,
										averaged_result_v);
		ROT_OUT <= conv_std_logic_vector(averaged_result_v, 32);		--gemittelten Datenwert auf Port legen
	else
		ROT_OUT <= bitmuster_v;		--einzelnen (nicht gemittelten) Datenwert auf Port legen
	end if;

	--PROD_ID Register, 16bit binary format
 	bitmuster_v := x"00000000";					--alles auf '0' setzen
	bitmuster_v(15 downto 0) := prod_id_reg_s(15 downto 0);	--den Datenwert einkopieren
 	PROD_ID <= bitmuster_v;		--Datenwert auf Port legen

	--STATUS Register, 16bit binary format
 	bitmuster_v := x"00000000";					--alles auf '0' setzen
	bitmuster_v(15 downto 0) := status_reg_s(15 downto 0);	--den Datenwert einkopieren
 	STATUS <= bitmuster_v;		--Datenwert auf Port legen

	--ERROR Register, 16bit binary format
 	bitmuster_v := x"00000000";					--alles auf '0' setzen
	bitmuster_v(15 downto 0) := error_reg_s(15 downto 0);	--den Datenwert einkopieren
 	ERROR <= bitmuster_v;		--Datenwert auf Port legen

end if;
end process;



--! @brief Erzeugung des SPI-SCLK-Taktes bzw. des reduzierten Systemclocks
--! @details Eine volle SCLK-Periode wird in x Teilen getaktet, damit ein sauberer 50%-DutyCycle zwischen Daten und SCLK erreicht wird. \n
--! Folglich läuft die SCLK-Takterzeugung bzw. der reduz. Systemclock mit adis_sclk_x_tel_s = 2 * 500kHz = 1MHz ---> daraus also SCLK = 500kHz. \n
--! Das Signal adis_sclk_x_tel_s wirkt als asynchrones enable-Signal für Prozesse.
spi_adis_sclk_x_tel: process(RESET_n, CLK)

variable count_4tel_v:	integer range 0 to spi_serial_clock_div_g;
begin
if RESET_n = '0' then
	adis_sclk_x_tel_s <= '0';
	count_4tel_v := 0;
elsif (CLK'EVENT and CLK = '1') then -- voller 50MHz-Systemclock !!!
	if count_4tel_v < (spi_serial_clock_div_g -1) then
		count_4tel_v := count_4tel_v + 1;
		adis_sclk_x_tel_s <= '0';
	else
		count_4tel_v := 0;
		adis_sclk_x_tel_s <= '1';
	end if;
end if;
end process;

end arch_adis16209_control;