--! @file    	adis16305_control.vhd
--! @brief   	VHDL-Modul, second processed file by DoxyGen 1.8.10
--! @details \n
--! Der ADIS16305 erfasst selbstständig Messwerte und stellt diese in Registern im Raw-Format über ein SPI-Interface zur Verfügung. \n
--! Der ADIS wird in einer fixen Konfiguration betrieben:
--! - nach Power-Up wird die Produkt-ID abgefragt und ein Selbsttest ausgelöst
--! - der aktuelle Status wird über das DIAG_STAT-Register ausgelesen
--!
--! <b>Wesentliche Funktionen</b>\n
--! - 2-stufige Synchronisation der IN-Ports mit dem Systemtakt, um ein stabiles VHDL-Verhalten bei Leseoperationen zu gewährleisten (Prozess "adis_port_sync")
--! - reduzierter Systemclock für SPI-Interface, SPI-SCLK = 500kHz, Erzeugung des asynchr. Enable-Signals (Prozess "spi_adis_sclk_x_tel")
--! - Auslesen der Messwerte, Aufbereitung, Mittelwertbildung und synchronisierte Port-Ausgabe (Prozess "outport_sync") \n
--!		- sobald das Data-Ready-Signal aktiv ist (DIO1 = high-aktiv), wird ein Burst-Read gestartet und alle Raw-Datenwerte werden in einem Durchlauf vom ADIS eingelesen \n
--!		- Beschneidung der 16bit-ADIS-Raw-Messdaten auf das reale Datenformat (12/14 bits, ohne ND/EA-bits), bei Bedarf Formatierung im 2's complement \n
--!		Messwerte in welchen Einheiten tatsächlich verwendet werden.
--!
--! - Gesamtsteuerung (Prozess "adis_main_statemachine") \n
--!		- Abschnitt Initialisierung, Selbsttest, Statusabfrage und Parameterkonfiguration \n
--!		- Abschnitt RX SINGLE Data Word from ADIS \n
--!
--!
--! @note
--! - Das beobachtete Verhalten weicht manchmal  vom Datenblatt (bzw. der Erwartungshaltung) ab, Details siehe kommentiertes Datenblatt.
--! - Es wird hier bewusst auf eine Nios-Steuerung über ein Command-Port verzichtet, da der ADIS in einer fixen Konfiguration läuft.
--!   Bei Bedarf muss ggf. das Command-Port hinzugefügt bzw. aktiviert werden.
--!
--! \n@verbatim
--! Modul-Interface-Register
--! Messwert-Ausgabe-Register mit den Messwerten (Inhalt siehe Datenblatt, Reihenfolge entsprechend der Ausgabe im Burst-Read-Mode, bereits formatiert für 32bit-Nios-Lesen )
--! 	#1	SUPPLY_OUT  : real 12bit offset binary format	
--! 	#2	GYRO_OUT    : real 14bit 2's complement format
--! 	#3	XACCL_OUT   : real 14bit 2's complement format
--! 	#4	YACCL_OUT   : real 14bit 2's complement format
--! 	#5	ZACCL_OUT   : real 14bit 2's complement format
--! 	#6	TEMP_OUT    : real 12bit 2's complement format
--! 	#7	PITCH_OUT   : real 13bit 2's complement format
--! 	#8	ROLL_OUT    : real 13bit 2's complement format
--! 	#9	AUX_ADC     : real 12bit offset binary format
--!
--! Production-ID-Register
--! PROD_ID          : real 16bit binary format, Silizium-ID 0x3FB1 (=16305)
--!
--! Diagnose-Status-Register
--! DIAG_STAT        : real 16bit binary format
--!
--! zusätzliches eigenes Fehler-Flag-Register
--! ERROR            : real 16bit binary format, "OR"-Mapping der einzelnen Fehlerflags (EA-bits) der Datenblöcke (#x) in ein Error-Flag-Register
--!  _________________________________________________________________________________________________________________________________
--! |     15|       |       |       |       |       |      9|       8||      7|      6|      5|      4|      3|      2|      1|      0|
--! |       |       |       |       |       |       |       |        ||       |       |       |       |       |       |       |       |
--! |  ---  |  ---  |  ---  |  ---  |  ---  |  ---  |   #9  |   #8   ||  #7   |  #6   |   #5  |  #4   |  #3   |  #2   |  #1   |  ---  |
--! |_______|_______|_______|_______|_______|_______|_______|________||_______|_______|_______|_______|_______|_______|_______|_______|
--!
--! @endverbatim \n
--!
--! compilation : QuartusII, V15.0.1, Build 150 \n
--! simulation : ModelSim, V10.3d rev. 2014.10 \n
--! 
--! @author  JMayr
--! @version V1.0
--! @date    02.09.2015
--! 
--! \n@par History:
--! @verbatim
--! V1.0	first release
--!		JMayr, 02.09.2015
--! @endverbatim
--!
--! \n@todo  nichts eingetragen
--!
--! @bug  nichts eingetragen
--!
--! @brief
--! <h2><center>&copy; COPYRIGHT 2011 UniBwM ETTI WE 4</center></h2>
--!
--! @defgroup VHDL VHDL

--! Use Standard Library
LIBRARY IEEE;

--! Use Standard Logic Library
USE IEEE.STD_LOGIC_1164.ALL;

--! Use Standard Arithmetic Library
USE IEEE.STD_LOGIC_ARITH.ALL;     -- für conv functions erforderlich

-- --! Use Standard Logic Unsigned Library
-- USE IEEE.STD_LOGIC_unsigned.ALL;  -- für conv functions erforderlich

--! Use Standard Logic Signed Library
USE IEEE.STD_LOGIC_signed.ALL;  -- alternativ zur unsigned-Lib...

-- --! Use Numeric Library
-- USE ieee.numeric_bit.ALL;
-- --! Use Numeric Library
-- USE ieee.numeric_std.ALL;

-- LIBRARY modelsim_lib;
-- use modelsim_lib.util.all;

--! @ingroup VHDL
--! @brief Ansteuerung ADIS16305
ENTITY adis16305_control is
GENERIC(
   spi_serial_clock_div_g      	: integer := 50;    	--! Teilerfaktor für operativen SPI-SCLK-Betrieb mit 500kHz (= 2µs): 50MHz / (2 x 500kHz) = 50 => 1MHz reduz. Clock (wg. adis_sclk_x_tel_s)
   dynamic_range_setting_g     	: integer := 4; 		--! Bits SENS_AVG[10:8] ---> '100' = +-300°/sec (default), '010' = +-150°/sec, '001' = +-75°/sec
   digital_filtering_setting_g	: integer := 2;  		--! Bits SENS_AVG[2:0]  ---> '010' ---> M = 2 ---> 2**2 = 4 filter taps (default)
   enable_data_averaging_g     	: integer := 1      	--! Averaging 8-level-Buffer für Messwerte: enable = 1, disable = 0 (wg. QSys nur Integer-Angabe)
    );
PORT(
   COMMAND         : in std_logic_vector(15 downto 0); --! 16-bit-Command-Port (hier nicht verwendet, weil ADIS in fester Konfiguration betrieben wird)
    
   -- ADIS16305-Daten- und Messwerte, für NIOS-Lesen bereits in 32bit-Standard-Wortbreite, ggf. bereits im 2's-Complement Format
   SUPPLY_OUT  : out std_logic_vector(31 downto 0);    	--! real 12bit offset binary format 
   GYRO_OUT        : out std_logic_vector(31 downto 0);  --! real 14bit 2's complement format
   PROD_ID     : out std_logic_vector(31 downto 0);    	--! real 16bit binary format, Silizium-ID 0x3FB1 (=16305)
   DIAG_STAT   : out std_logic_vector(31 downto 0);    	--! real 16bit binary format

   --eigenes Fehlerregister
   ERROR           : out std_logic_vector(31 downto 0);  --! real 16bit binary format, Mapping der Fehlerflags (EA-bit) der einzelnen Datenblöcke
    
   -- ADIS16305-Interface
   ADIS16305_CS_n          : OUT std_logic;              --! low-aktives ADIS16305-Chip-Select (In-Pin "CS")
   ADIS16305_DIO1          : IN std_logic;               --! high-aktives "Data-Ready"-Signal vom ADIS16305 (Out-Pin "DIO1")

   CLK         : IN std_logic;                           --! System-Clock
   RESET_n     : IN std_logic                            --! System-Reset
   );
END adis16305_control;


--! @brief Architecture für adis16305_control
ARCHITECTURE arch_adis16305_control of adis16305_control is

--definition main_state type
type main_state_type is ( 														--!States für Steuerung und Ablauf
	INIT_A_ST, INIT_B_ST, INIT_C_ST, INIT_D_ST, INIT_E_ST, INIT_F_ST, INIT_G_ST, INIT_H_ST, INIT_I_ST, INIT_J_ST, WAIT_DATA_READY_ST, START_BURST_READ_ST,
	GET_DATA_WORD_1_ST, GET_DATA_WORD_2_ST, GET_DATA_WORD_3_ST, GET_DATA_WORD_4_ST, GET_DATA_WORD_5_ST, GET_DATA_WORD_6_ST,
	GET_BURST_DATA_1_ST, GET_BURST_DATA_2_ST, GET_BURST_DATA_3_ST, GET_BURST_DATA_4_ST, GET_BURST_DATA_5_ST, GET_BURST_DATA_6_ST, GET_BURST_DATA_7_ST,
	TX_TO_ADIS_1_ST, TX_TO_ADIS_2_ST, TX_TO_ADIS_3_ST, TX_TO_ADIS_4_ST, TX_TO_ADIS_5_ST,
	ERROR_ST
	);
signal main_state_st 			: main_state_type;		--! für statemachine
signal next_main_state_st 		: main_state_type;		--! für statemachine, zum Merken des nächsten states

--Definition eines Averaging 8-level buffer types
type type_averaging_buffer is array (7 downto 0) of integer;		--! Datentyp für zusätzliche Messwert-Averaging-Funktion mit fixem 16-bit-Buffer für 8 Durchläufe (keine integer-Range-Angabe wg. Vorzeichen)
signal averaging_buffer_supply_out_s	: type_averaging_buffer;	--! Averaging-Buffer für SUPPLY_OUT


--! @brief VHDL-Procedure: It seems that this causes problems in documenting VHDL-Generics
--! @details 	The first processed file in the DoxyGen file list shows all Generics. \n
--!				But the second file does skip the Generics !
--!				Just put the procedure statements in the first processed file into comments ---> Generics in both files are shown
--! @param  type_of_result_v : aktueller zu mittelnder ADIS-Messwert
--! @param  averaging_buffer_<adis-register_out_s> : jeweils der Average-Buffer für den aktuell zu mittelnden Messwert
--! @retval local_averaged_result_v : errechneter Mittelwert
procedure proc_result_averaging (variable type_of_result_v					: integer;
											variable bitmuster_v							: in std_logic_vector;
											signal averaging_buffer_supply_out_s	: inout type_averaging_buffer;	--case type_of_result_v = 1
											signal averaging_buffer_gyro_out_s		: inout type_averaging_buffer;	--case type_of_result_v = 2
											signal averaging_buffer_xaccl_out_s		: inout type_averaging_buffer;	--case type_of_result_v = 3
											signal averaging_buffer_yaccl_out_s		: inout type_averaging_buffer;	--case type_of_result_v = 4
											signal averaging_buffer_zaccl_out_s		: inout type_averaging_buffer;	--case type_of_result_v = 5
											signal averaging_buffer_temp_out_s		: inout type_averaging_buffer;	--case type_of_result_v = 6
											signal averaging_buffer_pitch_out_s		: inout type_averaging_buffer;	--case type_of_result_v = 7
											signal averaging_buffer_roll_out_s		: inout type_averaging_buffer;	--case type_of_result_v = 8
											signal averaging_buffer_aux_adc_s		: inout type_averaging_buffer;	--case type_of_result_v = 9
											variable local_averaged_result_v			: out integer) is

begin
case type_of_result_v is
	when 1 => 		--supply_out
		averaging_buffer_supply_out_s(7) 	<= averaging_buffer_supply_out_s(6);
		averaging_buffer_supply_out_s(6) 	<= averaging_buffer_supply_out_s(5);
		averaging_buffer_supply_out_s(5) 	<= averaging_buffer_supply_out_s(4);
		averaging_buffer_supply_out_s(4) 	<= averaging_buffer_supply_out_s(3);
		averaging_buffer_supply_out_s(3) 	<= averaging_buffer_supply_out_s(2);
		averaging_buffer_supply_out_s(2) 	<= averaging_buffer_supply_out_s(1);
		averaging_buffer_supply_out_s(1) 	<= averaging_buffer_supply_out_s(0);
		averaging_buffer_supply_out_s(0) 	<= conv_integer(bitmuster_v);
		local_averaged_result_v := (	averaging_buffer_supply_out_s(0) +
												averaging_buffer_supply_out_s(1) +
												averaging_buffer_supply_out_s(2) +
												averaging_buffer_supply_out_s(3) +
												averaging_buffer_supply_out_s(4) +
												averaging_buffer_supply_out_s(5) +
												averaging_buffer_supply_out_s(6) +
												averaging_buffer_supply_out_s(7) ) / 8;

	when others => null;
end case;
end procedure proc_result_averaging;




--ADIS-Config-Konstanten
constant adis_powerup_div_factor_c				:integer := 2**18;	--! Teilerfaktor für die Power-Up-Wartezeit, mind. 250ms (reduz. Systemclock 1MHz(1µs) * 2**18 => ca. 260ms)
constant adis_prod_id_reg_read_mask_c			:std_logic_vector(15 DOWNTO 0) := "0101011000000000";	--! Reg. PROD_ID, read lower/upper Byte, ADR = 0x56, Bitmuster "0_1010110_0000_0000", Silizium-ID 0x3FB1

--**************************************************************************
--**************************************************************************
begin	--of architecture ****************************************************
--**************************************************************************
--**************************************************************************

--HW-ADIS16305-Signal-to-Port Assignment
ADIS16305_DIN 	<= adis_din_s;
ADIS16305_CS_n <= adis_cs_n_s;
ADIS16305_SCLK <= adis_sclk_s;


--! @brief Messwert-Aufbereitung, Mittelwertbildung und synchronisierte Port-Ausgabe
--! @details 	Die 16bit-ADIS-Raw-Messdaten werden auf das reale Datenformat (12/14 bits, ohne ND/EA-bits) beschnitten und bei Bedarf im 2's complement formatiert. \n
--!				Bei aktiviertem Averaging-QSys-Parameter werden die Messwerte dann einer gleitenden 8-Level-Mittelwertbildung unterzogen. \n
--!				Die Ausgabe erfolgt synchron mit dem "reduzierten" Systemtakt von 1MHz somit als "signed integer"-Zahlenwert.
--!				Eine evtl. Umrechnung dieser Zahlenwerte in andere physikalische Einheiten (z.B. "g", "V" oder "°/sec") erfolgt in der Applikationsoftware
outport_sync: process  (RESET_n, CLK, adis_sclk_x_tel_s,
								supply_out_raw_reg_s, gyro_out_raw_reg_s, xaccl_out_raw_reg_s, yaccl_out_raw_reg_s, zaccl_out_raw_reg_s,
								temp_out_raw_reg_s, pitch_out_raw_reg_s, roll_out_raw_reg_s, aux_adc_raw_reg_s, diag_stat_reg_s, error_reg_s)
							
variable bitmuster_v			: std_logic_vector(31 downto 0);	--Zwischenvariable zum temporären "Bit-Mixen"
variable type_of_result_v 	: integer range 0 to 15;			--Kennzeichnung der Art des Messwertes, damit in Average-Procedure die richtige Mittelwertbildung erfolgt
variable averaged_result_v	: integer;								--errechnetes Mittelwertergebnis 

begin
if RESET_n = '0' then
	bitmuster_v := x"00000000";
elsif adis_sclk_x_tel_s = '0' then           --asynchr. enable signal für "reduzierten Systemclock"
	null;
elsif CLK'EVENT and CLK = '1' then				--reduzierter Clock, de facto 1 MHz
	--für SUPPLY_OUT Register, real 12bit offset binary format
 	bitmuster_v := x"00000000";					--alles auf '0' setzen
	bitmuster_v(11 downto 0) := supply_out_raw_reg_s(11 downto 0);	--nur den realen Zahlenwert einkopieren
	type_of_result_v := 1;							--Kennzeichnung für Mittelwert-Prozedur setzen
	if (enable_data_averaging_g = 1) then
		proc_result_averaging (	type_of_result_v,
										bitmuster_v,
										averaging_buffer_supply_out_s,
										averaging_buffer_gyro_out_s,
										averaging_buffer_xaccl_out_s,
										averaging_buffer_yaccl_out_s,
										averaging_buffer_zaccl_out_s,
										averaging_buffer_temp_out_s,
										averaging_buffer_pitch_out_s,
										averaging_buffer_roll_out_s,
										averaging_buffer_aux_adc_s,
										averaged_result_v);
		SUPPLY_OUT <= conv_std_logic_vector(averaged_result_v, 32);		--gemittelten Datenwert auf Port legen
	else
		SUPPLY_OUT <= bitmuster_v;		--einzelnen (nicht gemittelten) Datenwert auf Port legen
	end if;
end if;
end process;



--! @brief Erzeugung des SPI-SCLK-Taktes bzw. des reduzierten Systemclocks
--! @details Eine volle SCLK-Periode wird in x Teilen getaktet, damit ein sauberer 50%-DutyCycle zwischen Daten und SCLK erreicht wird. \n
--! Folglich läuft die SCLK-Takterzeugung bzw. der reduz. Systemclock mit adis_sclk_x_tel_s = 2 * 500kHz = 1MHz ---> daraus also SCLK = 500kHz. \n
--! Das Signal adis_sclk_x_tel_s wirkt als asynchrones enable-Signal für Prozesse.
spi_adis_sclk_x_tel: process(RESET_n, CLK)

variable count_4tel_v:	integer range 0 to spi_serial_clock_div_g;
begin
if RESET_n = '0' then
	adis_sclk_x_tel_s <= '0';
	count_4tel_v := 0;
elsif (CLK'EVENT and CLK = '1') then -- voller 50MHz-Systemclock !!!
	if count_4tel_v < (spi_serial_clock_div_g -1) then
		count_4tel_v := count_4tel_v + 1;
		adis_sclk_x_tel_s <= '0';
	else
		count_4tel_v := 0;
		adis_sclk_x_tel_s <= '1';
	end if;
end if;
end process;

end arch_adis16305_control;